@@ -59,6 +59,11 @@ module Generics.SOP.NP
   , sequence'_POP
   , sequence_NP
   , sequence_POP
+    -- * Catamorphism and anamorphism
+  , cata_NP
+  , ccata_NP
+  , ana_NP
+  , cana_NP
   ) where
 
 #if !(MIN_VERSION_base(4,8,0))
@@ -66,6 +71,8 @@ import Control.Applicative
 #endif
 import Data.Proxy (Proxy(..))
 
+import Control.DeepSeq (NFData(..))
+
 import Generics.SOP.BasicFunctors
 import Generics.SOP.Classes
 import Generics.SOP.Constraint
@@ -109,6 +116,11 @@ deriving instance All (Show `Compose` f) xs => Show (NP f xs)
 deriving instance All (Eq   `Compose` f) xs => Eq   (NP f xs)
 deriving instance (All (Eq `Compose` f) xs, All (Ord `Compose` f) xs) => Ord (NP f xs)
 
+-- | @since 0.2.5.0
+instance All (NFData `Compose` f) xs => NFData (NP f xs) where
+    rnf Nil       = ()
+    rnf (x :* xs) = rnf x `seq` rnf xs
+
 -- | A product of products.
 --
 -- This is a 'newtype' for an 'NP' of an 'NP'. The elements of the
@@ -129,6 +141,10 @@ deriving instance (Show (NP (NP f) xss)) => Show (POP f xss)
 deriving instance (Eq   (NP (NP f) xss)) => Eq   (POP f xss)
 deriving instance (Ord  (NP (NP f) xss)) => Ord  (POP f xss)
 
+-- | @since 0.2.5.0
+instance (NFData (NP (NP f) xss)) => NFData (POP f xss) where
+    rnf (POP xss) = rnf xss
+
 -- | Unwrap a product of products.
 unPOP :: POP f xss -> NP (NP f) xss
 unPOP (POP xss) = xss
@@ -497,3 +513,92 @@ sequence_POP :: (All SListI xss, Applicative f) => POP f xss -> f (POP I xss)
 sequence_NP   = hsequence
 sequence_POP  = hsequence
 
+-- * Catamorphism and anamorphism
+
+-- | Catamorphism for 'NP'.
+--
+-- This is a suitable generalization of 'foldr'. It takes
+-- parameters on what to do for 'Nil' and ':*'. Since the
+-- input list is heterogeneous, the result is also indexed
+-- by a type-level list.
+--
+-- @since 0.2.3.0
+--
+cata_NP ::
+     forall r f xs .
+     r '[]
+  -> (forall y ys . f y -> r ys -> r (y ': ys))
+  -> NP f xs
+  -> r xs
+cata_NP nil cons = go
+  where
+    go :: forall ys . NP f ys -> r ys
+    go Nil       = nil
+    go (x :* xs) = cons x (go xs)
+
+-- | Constrained catamorphism for 'NP'.
+--
+-- The difference compared to 'cata_NP' is that the function
+-- for the cons-case can make use of the fact that the specified
+-- constraint holds for all the types in the signature of the
+-- product.
+--
+-- @since 0.2.3.0
+--
+ccata_NP ::
+     forall c proxy r f xs . (All c xs)
+  => proxy c
+  -> r '[]
+  -> (forall y ys . c y => f y -> r ys -> r (y ': ys))
+  -> NP f xs
+  -> r xs
+ccata_NP _ nil cons = go
+  where
+    go :: forall ys . (All c ys) => NP f ys -> r ys
+    go Nil       = nil
+    go (x :* xs) = cons x (go xs)
+
+-- | Anamorphism for 'NP'.
+--
+-- In contrast to the anamorphism for normal lists, the
+-- generating function does not return an 'Either', but
+-- simply an element and a new seed value.
+--
+-- This is because the decision on whether to generate a
+-- 'Nil' or a ':*' is determined by the types.
+--
+-- @since 0.2.3.0
+--
+ana_NP ::
+     forall s f xs .
+     SListI xs
+  => (forall y ys . s (y ': ys) -> (f y, s ys))
+  -> s xs
+  -> NP f xs
+ana_NP uncons = go sList
+  where
+    go :: forall ys . SList ys -> s ys -> NP f ys
+    go SNil  _ = Nil
+    go SCons s = case uncons s of
+      (x, s') -> x :* go sList s'
+
+-- | Constrained anamorphism for 'NP'.
+--
+-- Compared to 'ana_NP', the generating function can
+-- make use of the specified constraint here for the
+-- elements that it generates.
+--
+-- @since 0.2.3.0
+--
+cana_NP ::
+     forall c proxy s f xs . (All c xs)
+  => proxy c
+  -> (forall y ys . c y => s (y ': ys) -> (f y, s ys))
+  -> s xs
+  -> NP f xs
+cana_NP _ uncons = go sList
+  where
+    go :: forall ys . (All c ys) => SList ys -> s ys -> NP f ys
+    go SNil  _ = Nil
+    go SCons s = case uncons s of
+      (x, s') -> x :* go sList s'
